unit bigint;
{$R-}
interface

type
  TDigits=array [0..0] of longword;
  PDigits=^TDigits;
  TBigInt=record
    sgn,len:integer;
    digit:PDigits;
  end;

procedure dmove(src,dst:PDigits; len:longword); register;  //zkopiruj len 32-bit. cifer ze src do dst
procedure derase(dst:PDigits; len:longword); register;  //nuluj len 32-bit. cifer na pozicii dst

function i_fromStr(s:string):TBigInt;
function i_toStr(a:TBigInt):string;
procedure i_add(a:TBigInt; var b:TBigInt);   {k cislu b pricti cislo a}
procedure i_sub(a:TBigInt; var b:TBigInt);   {od cisla b odecti cislo a}
procedure i_sum(a,b:TBigInt; var c:TBigInt); {secti a a b a uloz na misto c}
procedure i_diff(a,b:TBigInt; var c:TBigInt); {odecti b od a a uloz na misto c}
procedure i_addspecl(n:integer; a:TBigInt; var b:TBigInt); {pricte k b cislo a*B^n, (B=2^32), a,b>=0, pokud je b kratsi nez n, musi po nem nasledovat nuly az do delky n}
procedure i_addspecr(n:integer; a:TBigInt; var b:TBigInt); {k b pricte a*B^-n, (B=2^32), pred b musi byt n mist volnych, a,b>=0}

procedure shlmod(a,b:PDigits; j,m:longword); pascal; //b:=(a*2^j) mod (2^(32*m)+1); 0<=j<32*m; delka a je m+1; znici a
procedure shrmod(a,b:PDigits; j,m:longword); pascal; //b:=(a*2^(-j)) mod (2^(32*m)+1); 0<j<=32*m; delka a je m+1; znici a

procedure DFTmod(n,om,mdl,step:longword; a,w:PDigits); //N=2^n, omega=2^om, modulo=2^(32*mdl)+1=omega^(N/2)+1, step:=mdl+1; workspace needed N/2*(mdl+1)
procedure iDFTmod(n,om,mdl:longword; a,w:PDigits); //N=2^n, omega=2^om, modulo=2^(32*mdl)+1=omega^(N/2)+1, step:=mdl+1; workspace needed N/2*(mdl+1)
procedure iDFTmod_(n,om,mdl,step:longword; a,w:PDigits); //N=2^n, omega=2^om, modulo=2^(32*mdl)+1=omega^(N/2)+1, step:=mdl+1; workspace needed N/2*(mdl+1)

procedure mod_diff(a,b:PDigits; m:longword); register; //b:=2^(m*32)+1-a
procedure mod_mod(a,b:PDigits; m:longword); register; //b:=a mod (2^(m*32)+1), a<2^(2*m*32)
procedure mod_modp(a,b:PDigits; m,n:longword); register; //b:=a mod (2^(m*32)+1), delka a je m+n cislic, 0<n<m
procedure mod_modn(a,b:PDigits; m,n:longword); register; //b:=(-a) mod (2^(m*32)+1), delka a je m+n cislic, 0<n<m

procedure ss_interlace(a,b:PDigits; m,l:longword); register;   //1<=m<=16, l=delka polozek a v bytech, delka b bude ceil(3m2^m/32), +1 prac. mista
procedure ss_extract(a,b:PDigits; m:longword); register;    //v a je serie m-bitovych cisel a_i po bitech, za poslednim je 0, do b ulozi m 16-bit. cisel (a_i-a_{m+i}) mod 2^16, vyznamnych je pouze prvnich m bitu; 1<=m<=16

implementation

procedure dmove(src,dst:PDigits; len:longword); register;
asm
  push esi
  push edi
  mov esi,eax
  mov edi,edx
  rep movsd
  pop edi
  pop esi
end;

procedure derase(dst:PDigits; len:longword); register;
asm
  push edi
  mov edi,eax
  mov ecx,edx
  xor eax,eax
  rep stosd
  pop edi
end;

function i_fromStr(s:string):TBigInt;
var z,i,j,d,len,rem:integer;
    c:longword;
    a:TBigInt;
begin
  if s[1]='-' then begin a.sgn:=-1; z:=1; end else begin a.sgn:=1; z:=0; end;
  d:=length(s);
  len:=(d-z) div 8;
  if (d-z) and $07>0 then inc(len);
  getmem(a.digit,4*len);
  d:=d-7;
  for i:=0 to len-1 do begin
    c:=0;
    if d<1+z then begin rem:=d+7-z; d:=1+z; end else rem:=8;
    for j:=1 to rem do begin
      if ord(s[d])<=ord('9') then c:=(c shl 4)+ord(s[d])-ord('0')
        else c:=(c shl 4)+ord(s[d])-ord('a')+10;
      inc(d);
    end;
    d:=d-16;
    a.digit[i]:=c;
  end;
  i:=len-1;
  while (i>=0) and (a.digit[i]=0) do dec(i);
  a.len:=i+1;
  if i<0 then a.sgn:=0;
  result:=a;
end;

function i_toStr(a:TBigInt):string;
var i,j,l,ll:integer;
    k:longword;
begin
  if a.sgn=0 then begin result:='0'; exit end;
  ll:=a.len; if ll>8 then ll:=8;
  for i:=0 to ll-1 do begin
    k:=a.digit[i];
    for j:=1 to 8 do begin
      l:=k and $0f;
      if l<10 then l:=l+ord('0') else l:=l-10+ord('a');
      result:=chr(l)+result;
      k:=k shr 4;
    end;
  end;
  i:=1;
  while (i<length(result)) and (result[i]='0') do inc(i);
  if i>1 then result:=copy(result,i,length(result)-i+1);
  if a.len>8 then result:='...'+result;
  if a.sgn=-1 then result:='-'+result;
end;

procedure i_add(a:TBigInt; var b:TBigInt);
var i,m,n:integer;
begin
  if a.sgn<>b.sgn then begin
    i:=a.len-1;
    while (i>=0) and (a.digit[i]=0) do dec(i);
    a.len:=i+1;
    i:=b.len-1;
    while (i>=0) and (b.digit[i]=0) do dec(i);
    b.len:=i+1; if b.len=0 then b.sgn:=0;
  end;
  if a.len=0 then exit;
  if b.sgn=0 then begin
    asm                            {je-li b=0, kopirujeme na misto b cislo a}
      push esi
      push edi
      mov edx,b
      mov ecx,a.sgn
      mov TBigInt[edx].sgn,ecx
      mov ecx,a.len
      mov TBigInt[edx].len,ecx
      mov esi,a.digit
      mov edi,TBigInt[edx].digit
      rep movsd
      pop edi
      pop esi
    end;
    exit;
  end;
  m:=b.len;
  if a.len<=m then begin
    if a.sgn=b.sgn then begin
      asm
        push esi
        push edi
        mov edx,b
        mov edi,TBigInt[edx].digit
        mov esi,a.digit
        mov ecx,a.len          {pocet cislic kratsiho cisla}
        clc                    {vynulovani prenosu}
@loop1:
        lodsd                  {nacti cislici cisla a do eax}
        adc [edi],eax          {pricti k cislici b cislici a}
        lea edi,[edi+4]        {posun ukazatel}
        dec ecx
        jnz @loop1            {opakuj pres cele a}
        jnc @done              {pokud neni prenos, koncime}
        mov ecx,TBigInt[edx].len
        sub ecx,a.len
        jz @carry              {pokud se rovanji delky a a b, mame prenos mimo vysledek}
@loop2:
        add [edi],1            {pricti 1=prenos}
        lea edi,[edi+4]
        jnc @done              {pokud neni dalsi prenos, koncime}
        dec ecx
        jnz @loop2            {pokracujeme v sireni prenosu}
@carry:                        {nastane prenos mimo vysledek}
        inc TBigInt[edx].len   {zvetsi velikost b o 1}
        mov [edi],1            {uloz 1 jako nejvyssi cislici vysledku}
@done:
        pop edi
        pop esi
      end;
    end else begin
      n:=a.len;
      if m=n then begin
        i:=n-1;
        while (i>=0) and (b.digit[i]=a.digit[i]) do begin b.digit[i]:=0; dec(i); end;  {hledej kde se cisla shoduji}
        if i<0 then begin b.len:=0; b.sgn:=0; exit; end else begin
          m:=i+1; n:=m;
          if b.digit[i]<a.digit[i] then begin
            b.sgn:=a.sgn;
            asm
              push esi
              push edi
              mov edx,b
              mov edi,TBigInt[edx].digit
              mov esi,a.digit
              mov ecx,m
              clc
  @loop1:
              lodsd
              sbb eax,[edi]      {odecti a-b}
              stosd
              dec ecx
              jnz @loop1
              sub edi,4         {hledej kolik nul je na zacatku vysledku}
              xor eax,eax
              mov ecx,m
              std
              repe scasd
              cld
              inc ecx
              mov TBigInt[edx].len,ecx
              pop edi
              pop esi
            end;
            exit;
          end;
        end;
      end; {nyni je a mensi nez b (v abs. hodnote), v m je 'aktivni' delka b, v n je 'aktivni' delka a}
      asm
        push esi
        push edi
        mov edx,b
        mov edi,TBigInt([edx]).digit
        mov esi,a.digit
        mov ecx,n
        clc
@loop1:
        lodsd
        sbb [edi],eax
        lea edi,[edi+4]
        dec ecx
        jnz @loop1
        jnc @done
@loop2:
        sub [edi],1
        lea edi,[edi+4]
        jc @loop2
@done:
        mov ecx,m         {hledej kolik nul je na zacatku vysledku}
        mov edi,TBigInt([edx]).digit
        lea edi,[edi+4*ecx-4]
        xor eax,eax
        std
        repe scasd
        cld
        inc ecx
        mov TBigInt[edx].len,ecx
        pop edi
        pop esi
      end;
    end;
  end else
{a je delsi nez b}
  if a.sgn=b.sgn then begin
    asm
      push esi
      push edi
      mov edx,b
      mov edi,TBigInt[edx].digit
      mov esi,a.digit
      mov ecx,TBigInt[edx].len          {pocet cifer b}
      clc
@loop1:
      lodsd
      adc [edi],eax
      lea edi,[edi+4]
      dec ecx
      jnz @loop1
      jnc @rest
      mov ecx,a.len
      sub ecx,TBigInt[edx].len
@loop2:
      lodsd
      add eax,1          {pricti 1=prenos}
      stosd
      dec ecx
      jnc @copy
      jnz @loop2
@carry:             {nastane prenos mimo nase cislo}
      mov eax,a.len
      inc eax   {zvetsi velikost b o 1}
      mov [edi],1            {uloz 1 jako nejvyssi cislici vysledku}
      jmp @done
@rest:               {kopirovani zbytku cisla}
      mov ecx,a.len
      sub ecx,TBigInt[edx].len
@copy:
      rep movsd
      mov eax,a.len
@done:
      mov TBigInt[edx].len,eax
      pop edi
      pop esi
    end;
  end else begin
    asm
      push esi
      push edi
      mov edx,b
      mov edi,TBigInt[edx].digit
      mov esi,a.digit
      mov ecx,TBigInt[edx].len
      mov eax,a.sgn
      mov TBigInt[edx].sgn,eax
      clc
@loop1:
      lodsd
      sbb eax,[edi]
      stosd
      dec ecx
      jnz @loop1
      jnc @rest
      mov ecx,a.len
      sub ecx,TBigInt[edx].len
@loop2:
      lodsd
      sub eax,1
      stosd
      dec ecx
      jc @loop2
      jmp @copy
@rest:
      mov ecx,a.len
      sub ecx,TBigInt[edx].len
@copy:
      rep movsd
@done:
      sub edi,4         {hledej kolik nul je na zacatku vysledku}
      xor eax,eax
      mov ecx,a.len
      std
      repe scasd
      cld
      inc ecx
      mov TBigInt[edx].len,ecx
      pop edi
      pop esi
    end;
  end;
end;

procedure i_sub(a:TBigInt; var b:TBigInt);
begin
  a.sgn:=-a.sgn;
  i_add(a,b);
end;

procedure i_sum(a,b:TBigInt; var c:TBigInt);
var i:integer;
    d:TBigInt;
begin
  if a.sgn<>b.sgn then begin
    i:=a.len-1;
    while (i>=0) and (a.digit[i]=0) do dec(i);
    a.len:=i+1;
    i:=b.len-1;
    while (i>=0) and (b.digit[i]=0) do dec(i);
    b.len:=i+1;
  end;
  if b.len>a.len then begin d:=a; a:=b; b:=d; end;
  if a.len=0 then begin c.sgn:=0; c.len:=0; exit; end;
  if b.len=0 then begin
    asm                            {je-li b=0, kopirujeme a do c}
      push esi
      push edi
      mov edx,c
      mov ecx,a.sgn
      mov TBigInt[edx].sgn,ecx
      mov ecx,a.len
      mov TBigInt[edx].len,ecx
      mov esi,a.digit
      mov edi,TBigInt[edx].digit
      rep movsd
      pop edi
      pop esi
    end;
    exit;
  end;
  if a.sgn=b.sgn then begin
    asm
      push esi
      push edi
      mov edx,c
      mov edi,TBigInt[edx].digit
      mov eax,a.sgn
      mov TBigInt[edx].sgn,eax
      mov eax,a.len
      mov TBigInt[edx].len,eax
      mov esi,a.digit
      mov edx,b.digit
      mov ecx,b.len
      clc
@loop1:
      lodsd
      adc eax,[edx]
      stosd
      lea edx,[edx+4]
      dec ecx
      jnz @loop1
      jnc @rest
      mov ecx,a.len
      sub ecx,b.len
      jz @carry
@loop2:
      lodsd
      add eax,1
      stosd
      dec ecx
      jnc @copy
      jnz @loop2
@carry:
      mov edx,c
      inc TBigInt[edx].len
      mov [edi],1
      jmp @done
@rest:
      mov ecx,a.len
      sub ecx,b.len
@copy:
      rep movsd
@done:
      pop edi
      pop esi
    end;
  end else begin
    if a.len=b.len then begin
      i:=a.len-1;
      while (i>=0) and (b.digit[i]=a.digit[i]) do dec(i);   {hledej kde se cisla shoduji}
      if i<0 then begin c.len:=0; c.sgn:=0; exit; end;
      if a.digit[i]<b.digit[i] then begin d:=a; a:=b; b:=d; end;
      a.len:=i+1; b.len:=i+1;
    end;
    {a je vetsi nez b, jejich delky jsou nyni 'aktivni'}
    asm
      push esi
      push edi
      mov edx,c
      mov edi,TBigInt[eax].digit
      mov eax,a.sgn
      mov TBigInt[edx].sgn,eax
      mov eax,a.len
      mov TBigInt[edx].len,eax
      mov esi,a.digit
      mov edx,b.digit
      mov ecx,b.len
      clc
@loop1:
      lodsd
      sbb eax,[edx]
      stosd
      lea edx,[edx+4]
      dec ecx
      jnz @loop1
      jnc @rest
      mov ecx,a.len
      sub ecx,b.len
@loop2:
      lodsd
      sub eax,1
      stosd
      dec ecx
      jnc @copy
      jnz @loop2
@rest:
      mov ecx,a.len
      sub ecx,b.len
@copy:
      rep movsd
@done:
      sub edi,4         {hledej kolik nul je na zacatku vysledku}
      xor eax,eax
      mov ecx,a.len
      std
      repe scasd
      cld
      inc ecx
      mov edx,c
      mov TBigInt[edx].len,ecx
      pop edi
      pop esi
    end;
  end;
end;

procedure i_diff(a,b:TBigInt; var c:TBigInt);
begin
  b.sgn:=-b.sgn;
  i_sum(a,b,c);
end;

procedure i_addspecl(n:integer; a:TBigInt; var b:TBigInt);
begin
  if a.sgn=0 then exit;
  if n>=b.len then begin
    asm
      push esi
      push edi
      mov edx,b
      mov edi,TBigInt[edx].digit
      mov esi,a.digit
      mov eax,n
      lea edi,[edi+4*eax]
      mov ecx,a.len
      add eax,ecx
      rep movsd
      mov TBigInt[edx].len,eax
      mov eax,a.sgn
      mov TBigInt[edx].sgn,eax
      pop edi
      pop esi
    end;
  end else begin
    dec(b.len,n);
    inc(b.digit,n);
    i_add(a,b);
    dec(b.digit,n);
    inc(b.len,n);
  end;
end;

procedure i_addspecr(n:integer; a:TBigInt; var b:TBigInt);
var m:integer;
begin
  m:=a.len;
  if m>n then m:=n;
  asm
    push esi
    push edi
    mov edx,b
    mov ecx,m
    mov edi,TBigInt[edx].digit
    mov eax,n
    shl eax,2
    sub edi,eax
    mov esi,a.digit
    rep movsd
    mov ecx,n
    sub ecx,m
    xor eax,eax
    rep stosd
    pop edi
    pop esi
  end;
  if a.len>n then begin
    dec(a.len,n);
    inc(a.digit,n);
    i_add(a,b);
  end;
end;

procedure shlmod(a,b:PDigits; j,m:longword); pascal; //b:=(a*2^j) mod (2^(32*m)+1); 0<=j<32*m; delka a je m+1; znici a
asm
  push esi
  push edi
  push ebx
  std
  mov ecx,j           //shift a (j mod 32) positions to the left
  and ecx,$1f
  jz @nosh
  mov esi,a
  mov ebx,m
  lea esi,[esi+4*ebx]
  lodsd
@loop1:
  mov edx,eax
  lodsd
  shld edx,eax,cl
  mov [esi+8],edx
  dec ebx
  jnz @loop1
  shl eax,cl
  mov [esi+4],eax
@nosh:                //test which half is bigger - upper (z1) or lower (z0)
  mov edx,j
  shr edx,5
  mov ebx,m
  mov eax,a
  lea esi,[eax+4*ebx]   //end of upper half
  sub ebx,edx
  dec ebx
  lea edi,[eax+4*ebx]   //end of lower half
  jz @topcmp
  xor eax,eax
  mov ecx,ebx
  repe scasd
  jne @z0aez1
@topcmp:
  mov ecx,[edi]
  cmp ecx,[esi]
  ja @z0aez1
  jb @z0bz1
  mov edi,esi
  sub edi,4
  mov ecx,edx
  test ecx,ecx
  jz @z0aez1
  xor eax,eax
  repe scasd
  jne @z0bz1
@z0aez1:
  cld
  mov edi,b
  mov esi,a
  mov ebx,m
  sub ebx,edx
  lea esi,[esi+4*ebx]
  mov ecx,edx
  jecxz @jis0a    //cf=0
  xor edx,edx
@loop2a:
  mov eax,edx
  sbb eax,[esi]
  stosd
  lea esi,[esi+4]
  dec ecx
  jnz @loop2a
@jis0a:
  mov ecx,[esi]
  mov esi,a
  lodsd
  sbb eax,ecx
  stosd
  mov ecx,ebx
  dec ecx
  jz @done
@loop3a:
  lodsd
  sbb eax,edx
  stosd
  dec ecx
  jnz @loop3a
  jmp @done
@z0bz1:
  cld
  mov edi,b
  mov esi,a
  mov ebx,m
  sub ebx,edx
  lea esi,[esi+4*ebx]
  mov ecx,edx
  test ecx,ecx
  jnz @jnot0
  mov ecx,[esi]   //cf=0
  mov esi,a
  lodsd
  inc eax
  jmp @dosub
@jnot0:
  xor edx,edx
  mov eax,1
  sbb eax,[esi]
  stosd
  lea esi,[esi+4]
  dec ecx
  jz @jdone
@loop2b:
  mov eax,edx
  sbb eax,[esi]
  stosd
  lea esi,[esi+4]
  dec ecx
  jnz @loop2b
@jdone:
  mov ecx,[esi]
  mov esi,a
  lodsd
@dosub:
  sbb eax,ecx
  stosd
  mov ecx,ebx
  dec ecx
  jz @done1
@loop3b:
  mov eax,edx
  sbb eax,edx
  stosd
  dec ecx
  jnz @loop3b
@done1:
  inc ecx
  sbb ecx,edx
@done:
  mov [edi],ecx
  pop ebx
  pop edi
  pop esi
end;

procedure shrmod(a,b:PDigits; j,m:longword); pascal; //b:=(a*2^(-j)) mod (2^(32*m)+1); 0<j<=32*m; delka a je m+1; znici a
asm
  push esi
  push edi
  push ebx
  std
  mov edx,j           //j:=32*m-j
  mov ecx,m
  shl ecx,5
  sub ecx,edx
  mov j,ecx
  and ecx,$1f         //shift a (j mod 32) positions to the left
  jz @nosh
  mov esi,a
  mov ebx,m
  lea esi,[esi+4*ebx]
  lodsd
@loop1:
  mov edx,eax
  lodsd
  shld edx,eax,cl
  mov [esi+8],edx
  dec ebx
  jnz @loop1
  shl eax,cl
  mov [esi+4],eax
@nosh:                //test which half is bigger - upper (z2) or lower (z1)
  mov edx,j
  shr edx,5
  mov ebx,m
  mov eax,a
  lea esi,[eax+4*ebx]   //end of upper half
  sub ebx,edx
  dec ebx
  lea edi,[eax+4*ebx]   //end of lower half
  jz @topcmp
  xor eax,eax
  mov ecx,ebx
  repe scasd
  jne @z2bz1
@topcmp:
  mov ecx,[edi]
  cmp ecx,[esi]
  ja @z2bz1
@z2aez1:
  cld
  mov edi,b
  mov esi,a
  mov ebx,m
  sub ebx,edx
  lea esi,[esi+4*ebx]
  mov ecx,edx
  rep movsd
  mov eax,[esi]
  mov esi,a
  sub eax,[esi]
  stosd
  lea esi,[esi+4]
  mov ecx,ebx
  dec ecx
  xor eax,eax
  rep stosd
  jmp @done
@z2bz1:
  cld
  mov edi,b
  mov esi,a
  mov ebx,m
  sub ebx,edx
  lea esi,[esi+4*ebx]
  mov ecx,edx
  test ecx,ecx
  stc
  jz @jdone
  xor edx,edx
  mov eax,1
  add eax,[esi]
  stosd
  lea esi,[esi+4]
  dec ecx
  jz @jdone
@loop2b:
  mov eax,edx
  adc eax,[esi]
  stosd
  lea esi,[esi+4]
  dec ecx
  jnz @loop2b
@jdone:
  mov eax,[esi]
  mov esi,a
  adc eax,edx
  sub eax,[esi]
  lea esi,[esi+4]
  stosd
  mov ecx,ebx
  dec ecx
  jz @done1
@loop3b:
  mov eax,edx
  sbb eax,[esi]
  lea esi,[esi+4]
  stosd
  dec ecx
  jnz @loop3b
@done1:
  inc ecx
  sbb ecx,edx
@done:
  mov [edi],ecx
  pop ebx
  pop edi
  pop esi
end;

procedure i_diffmod(a,b,c:PDigits; m:longword); pascal; //c:=(a-b) mod (2^(m*32)+1); delka a,b je m+1; m<2^31
asm
  push esi
  push edi
  push ebx
  std                 //test which is bigger - a or b
  mov ecx,m
  mov esi,a
  lea esi,[esi+4*ecx]
  mov edi,b
  lea edi,[edi+4*ecx]
  inc ecx
  repe cmpsd
  mov edi,c
  mov ecx,m
  cld
  je @aeqb          //a=b
  mov esi,a
  mov edx,b
  jb @abb           //a<b
  inc ecx
@loop1:
  lodsd
  sbb eax,[edx]
  lea edx,[edx+4]
  stosd
  dec ecx
  jnz @loop1
  jmp @done
@abb:
  xor ebx,ebx
  inc ebx
@loop2:
  lodsd
  rcl ecx,1           //schovej carry
  add eax,ebx
  rcl ebx,1
  and ebx,1
  rcr ecx,1
  sbb eax,[edx]
  lea edx,[edx+4]
  stosd
  dec ecx
  jnz @loop2
  inc ebx
  mov eax,[esi]
  rcl ecx,1
  add eax,ebx
  rcr ecx,1
  sbb eax,[edx]
  mov [edi],eax
  jmp @done
@aeqb:
  inc ecx
  xor eax,eax
  rep stosd
@done:
  pop ebx
  pop edi
  pop esi
end;

procedure i_addmod(a,b:PDigits; m:longword);  pascal; //b:=(a+b) mod (2^(m*32)+1); delka a,b je m+1; m<2^31
asm
  push esi
  push edi
  mov esi,a
  mov edi,b
  mov ecx,m
  xor edx,edx
@loop1:
  mov eax,[edi]
  adc eax,[esi]
  lea esi,[esi+4]
  mov [edi],eax
  rcl ecx,1       //schovej carry
  or edx,eax
  rcr ecx,1
  lea edi,[edi+4]
  dec ecx
  jnz @loop1
  mov eax,[edi]
  adc eax,[esi]
  mov [edi],eax
  test eax,eax
  jz @done
  test eax,2
  jnz @ovfl
  test edx,edx
  jz @done
@ovfl:
  dec eax
  mov [edi],eax
  mov edi,b
@loop2:
  sub [edi],1
  lea edi,[edi+4]
  jc @loop2
@done:
  pop edi
  pop esi
end;

procedure DFTmod(n,om,mdl,step:longword; a,w:PDigits); //N=2^n, omega=2^om, modulo=2^(32*mdl)+1=omega^(N/2)+1, step:=mdl+1; workspace needed n/2*(mdl+1)
var i,j,jj,k,m,jom,st:longword;
begin
  if n>0 then begin
    dec(n);
    DFTmod(n,om shl 1,mdl,step shl 1,a,w);
    DFTmod(n,om shl 1,mdl,step shl 1,@a[step],w);
    m:=1 shl n;
    i_diffmod(a,@a[step],w,mdl);
    i_addmod(@a[step],a,mdl);
    jom:=0;
    st:=step shl 1;
    i:=step; k:=0; jj:=0;
    for j:=1 to m-1 do begin
      inc(jom,om);
      inc(k,st);
      inc(jj,mdl+1);
      shlmod(@a[k+step],@a[i],jom,mdl);
      i_diffmod(@a[k],@a[i],@w[jj],mdl);
      i_addmod(@a[k],@a[i],mdl);
      inc(i,step);
    end;
    asm             //presun w do druhe pulky vektoru
      push esi
      push edi
      push ebx
      mov esi,w
      mov edi,a
      mov eax,i
      lea edi,[edi+4*eax]
      mov ebx,m
      inc mdl
      mov edx,step
      sub edx,mdl
      shl edx,2
@loop1:
      mov ecx,mdl
      rep movsd
      add edi,edx
      dec ebx
      jnz @loop1
      pop ebx
      pop edi
      pop esi
    end;
  end;
end;

procedure iDFTmod_(n,om,mdl,step:longword; a,w:PDigits); //N=2^n, omega=2^om, modulo=2^(32*mdl)+1=omega^(N/2)+1, step:=mdl+1; workspace needed n/2*(mdl+1)
var i,j,jj,k,m,jom,st:longword;
begin
  if n>0 then begin
    dec(n);
    iDFTmod_(n,om shl 1,mdl,step shl 1,a,w);
    iDFTmod_(n,om shl 1,mdl,step shl 1,@a[step],w);
    m:=1 shl n;
    i_diffmod(a,@a[step],w,mdl);
    i_addmod(@a[step],a,mdl);
    jom:=0;
    st:=step shl 1;
    i:=step; k:=0; jj:=0;
    for j:=1 to m-1 do begin
      inc(jom,om);
      inc(k,st);
      inc(jj,mdl+1);
      shrmod(@a[k+step],@a[i],jom,mdl);
      i_diffmod(@a[k],@a[i],@w[jj],mdl);
      i_addmod(@a[k],@a[i],mdl);
      inc(i,step);
    end;
    asm             //presun w do druhe pulky vektoru
      push esi
      push edi
      push ebx
      mov esi,w
      mov edi,a
      mov eax,i
      lea edi,[edi+4*eax]
      mov ebx,m
      inc mdl
      mov edx,step
      sub edx,mdl
      shl edx,2
@loop1:
      mov ecx,mdl
      rep movsd
      add edi,edx
      dec ebx
      jnz @loop1
      pop ebx
      pop edi
      pop esi
    end;
  end;
end;

procedure iDFTmod(n,om,mdl:longword; a,w:PDigits); //N=2^n, omega=2^om, modulo=2^(32*mdl)+1=omega^(N/2)+1; workspace needed n/2*(mdl+1)
var i,m:longword;
begin
  if n=0 then exit;
  iDFTmod_(n,om,mdl,mdl+1,a,w);
  m:=1 shl n;
  for i:=1 to m do begin
    shrmod(a,w,n,mdl);      //0<n<=32*mdl
    asm                //presun vysledek z w do a, posun a
      push esi
      push edi
      mov ecx,mdl
      inc ecx
      mov esi,w
      mov edi,a
      mov eax,ecx
      shl eax,2
      add a,eax
      rep movsd
      pop edi
      pop esi
    end;
  end;
end;

procedure mod_diff(a,b:PDigits; m:longword); register; //b:=2^(m*32)+1-a
asm
  push esi
  push edi
  mov esi,eax
  mov edi,edx
  xor edx,edx
  mov eax,1
  sub eax,[esi]
  lea esi,[esi+4]
  stosd
  dec ecx
  jz @done
@loop:
  mov eax,edx
  sbb eax,[esi]
  lea esi,[esi+4]
  stosd
  dec ecx
  jnz @loop
@done:
  mov eax,1
  sbb eax,[esi]
  stosd
  pop edi
  pop esi
end;

procedure mod_mod(a,b:PDigits; m:longword); register; //b:=a mod (2^(m*32)+1), a<2^(2*m*32)
asm
  push esi
  push edi
  mov esi,eax
  mov edi,edx
  lea edx,[esi+4*ecx]
  clc
  push edi
@loop1:
  mov eax,[esi]
  sbb eax,[edx]
  lea esi,[esi+4]
  lea edx,[edx+4]
  stosd
  dec ecx
  jnz @loop1
  mov [edi],ecx
  pop edi
  jnc @done
  inc ecx
@loop2:
  add [edi],ecx
  lea edi,[edi+4]
  jc @loop2
@done:
  pop edi
  pop esi
end;

procedure mod_modp(a,b:PDigits; m,n:longword); register; //b:=a mod (2^(m*32)+1), delka a je m+n cislic, 0<n<m
asm
  push esi
  push edi
  mov esi,eax
  mov edi,edx
  lea edx,[esi+4*ecx]
  push edi
  mov eax,n
  sub ecx,eax
  mov n,ecx
  mov ecx,eax
@loop1:
  mov eax,[esi]
  sbb eax,[edx]
  lea esi,[esi+4]
  lea edx,[edx+4]
  stosd
  dec ecx
  jnz @loop1
  mov edx,n
@loop2:
  mov eax,[esi]
  sbb eax,ecx
  lea esi,[esi+4]
  stosd
  dec edx
  jnz @loop2
  mov [edi],ecx
  pop edi
  jnc @done
  inc ecx
@loop3:
  add [edi],ecx
  lea edi,[edi+4]
  jc @loop3
@done:
  pop edi
  pop esi
end;

procedure mod_modn(a,b:PDigits; m,n:longword); register; //b:=(-a) mod (2^(m*32)+1), delka a je m+n cislic, 0<n<m
asm
  push esi
  push edi
  mov esi,eax
  mov edi,edx
  lea edx,[esi+4*ecx]
  push edi
  mov eax,n
  sub ecx,eax
  mov n,ecx
  mov ecx,eax
@loop1:
  mov eax,[edx]
  sbb eax,[esi]
  lea esi,[esi+4]
  lea edx,[edx+4]
  stosd
  dec ecx
  jnz @loop1
  mov edx,n
@loop2:
  mov eax,ecx
  sbb eax,[esi]
  lea esi,[esi+4]
  stosd
  dec edx
  jnz @loop2
  mov [edi],ecx
  pop edi
  jnc @done
  inc ecx
@loop3:
  add [edi],ecx
  lea edi,[edi+4]
  jc @loop3
@done:
  pop edi
  pop esi
end;

procedure ss_interlace(a,b:PDigits; m,l:longword); register;   //1<=m<=16, l=delka polozek a v bytech, delka b bude ceil(3m2^m/32), +1 prac. mista
var msk,cnt:longword;
asm
  push esi
  push edi
  push ebx
  mov esi,eax
  mov edi,edx
  mov edx,1
  shl edx,cl
  mov cnt,edx
  mov edx,1
  shl edx,cl
  dec edx
  mov msk,edx
  mov ebx,ecx
  shl ebx,1
  add ebx,ecx
  mov eax,ebx
  shl eax,cl
  add eax,$1f
  shr eax,5
  mov ecx,eax       //ceil(3m2^m/32)
  xor eax,eax
  push edi
  rep stosd
  pop edi
@loop:
  mov eax,[esi]
  add esi,l
  and eax,msk
  xor edx,edx
  shld edx,eax,cl
  shl eax,cl
  or [edi],eax
  add ecx,ebx
  or [edi+4],edx
  mov eax,ecx
  and ecx,$1f
  shr eax,5
  lea edi,[edi+4*eax]
  dec cnt
  jnz @loop
  pop ebx
  pop edi
  pop esi
end;

procedure ss_extract(a,b:PDigits; m:longword); register;   //v a je serie m-bitovych cisel a_i po bitech, za poslednim je 0, do b ulozi m 16-bit. cisel (a_i-a_{m+i}) mod 2^16, vyznamnych je pouze prvnich m bitu; 1<=m<=16
var cnt1,cnt2:longword;
asm
  push esi
  push edi
  push ebx
  mov esi,eax
  mov edi,edx
  push edi
  mov edx,1
  shl edx,cl
  mov cnt1,edx
  mov cnt2,edx
  mov ebx,ecx
  shl ebx,1
  add ebx,ecx
  xor ecx,ecx
@loop1:
  mov eax,[esi]
  mov edx,[esi+4]
  shrd eax,edx,cl
  mov [edi],ax
  lea edi,[edi+2]
  add ecx,ebx
  mov eax,ecx
  and ecx,$1f
  shr eax,5
  lea esi,[esi+4*eax]
  dec cnt1
  jnz @loop1
  pop edi
@loop2:
  mov eax,[esi]
  mov edx,[esi+4]
  shrd eax,edx,cl
  mov dx,[edi]
  sub edx,eax
  mov [edi],dx
  lea edi,[edi+2]
  add ecx,ebx
  mov eax,ecx
  and ecx,$1f
  shr eax,5
  lea esi,[esi+4*eax]
  dec cnt2
  jnz @loop2
  pop ebx
  pop edi
  pop esi
end;

end.
